<?php
// Copyright 1999-2017. Plesk International GmbH. All rights reserved.
/**
 * Script for convertation of localization files from SMB to Plesk format
 *
 * Verification:
 * grep -R "=>" ../../resources/languages/en-US/ | grep -v .svn | wc -l
 * cat en-US.php | grep "=>" | wc -l
 */

/**
 * Return help message
 *
 * @return string
 */
function getHelp()
{
    $script = $GLOBALS['argv'][0];
    return <<<HELP
Usage: $script -d localeDirectory [-o outputDirectory]
-d localeDirectory  SMB locale storage directory
-o outputDirectory  output directory path (optional, current directory is used by default)
Example of usage: $script -d application/resources/languages -o locales

See http://wiki.plesk.ru/wiki/SMB_Translation#Interaction_with_the_localization_team for details

HELP;
}

/**
 * Return script options
 *
 * @return array
 */
function getOptions()
{
    $options = getopt('d:o:');
    if (false === $options || !array_key_exists('d', $options)) {
        throw new Exception(getHelp());
    } else {
        if (!array_key_exists('o', $options)) {
            $options['o'] = '.';
        }
    }
    return $options;
}

/**
 * Run utility
 *
 */
function run()
{
    $options = getOptions();
    $languagesPath = rtrim($options['d'], DIRECTORY_SEPARATOR);
    $outputPath = rtrim($options['o'], DIRECTORY_SEPARATOR);

    if (!is_dir($languagesPath)) {
        throw new Exception("SMB localization directory '$languagesPath' does not exist\n");
    }

    $localeCode = basename($languagesPath);

    echo "Preparing locale: $localeCode\n";
    echo "Source directory: $languagesPath\n";

    $originalDir = getcwd();
    chdir($languagesPath);

    $translations = array($localeCode => array());

    $files = new RecursiveIteratorIterator(new RecursiveDirectoryIterator('.'),
        RecursiveIteratorIterator::SELF_FIRST);

    foreach ($files as $fileName => $file) {
        if (false !== strpos($fileName, '.svn') || is_dir($fileName)) {
            continue;
        }

        echo "File: $fileName\n";

        $sectionName = str_replace('.php', '', $fileName);
        $sectionName = str_replace('.' . DIRECTORY_SEPARATOR, '', $sectionName);
        $sectionName = str_replace(DIRECTORY_SEPARATOR, '.', $sectionName);

        $translations[$localeCode][$sectionName] = getSectionContent("$fileName", $sectionName);
    }

    chdir($originalDir);

    output($translations, $outputPath);
}

/**
 * Get section content
 *
 * @param string $sectionFile
 * @param string $sectionName
 * @return array
 */
function getSectionContent($sectionFile, $sectionName)
{
    $lines = file($sectionFile);
    $messages = array();

    foreach ($lines as $line) {
        if (false === strpos($line, '=>')) {
            continue ;
        }

        $line = preg_replace('/^([^\'"]*[\'"])/', "\\1$sectionName.", $line);

        $messages[] = $line;
    }

    return $messages;
}

/**
 * Create set of Plesk localization files
 *
 * @param array $translations
 * @param string $outputDir
 */
function output($translations, $outputDir)
{
    if (!is_dir($outputDir)) {
        throw new Exception("Output directory '$outputDir' does not exist\n");
    }

    foreach ($translations as $lang => $sections) {
        $content = render($sections);
        $fileName = $outputDir . DIRECTORY_SEPARATOR . $lang . ".php";
        file_put_contents($fileName, $content);
        echo "Locale $lang have been prepared.\n";
        echo "File: $fileName\n";
    }

}

/**
 * Return locale file content
 *
 * @param array $sections
 * @return string
 */
function render($sections)
{
    $date = date(DATE_ATOM);

    $header = <<<PHP
<?php

// generation date: $date

\$GLOBALS['lmsg_arr'] += array (

PHP;
    $content = '';

    $footer = <<<PHP
);
PHP;

    foreach ($sections as $section) {
        $sectionContent = rtrim(join($section));

        if (preg_match('/[^,]$/', $sectionContent)) {
            $sectionContent = $sectionContent . ",";
        }

        $content .= $sectionContent . "\n\n";
    }

    return $header . $content . $footer;
}

try {
    run();
} catch (Exception $e) {
    echo $e->getMessage();
    exit(1);
}
