// Copyright 1999-2017. Plesk International GmbH. All rights reserved.

define([
    './error-dialog',
    './wait',
    './view',
], function (ErrorDialog, Wait, view) {
    return Class.create({

        initialize: function (config) {
            this._locale = config.locale;
            var confirmationDialogLocale = config.confirmationDialogLocale;
            this._errorDialogLocale = config.errorDialogLocale;

            this._list = Jsw.getComponent('files-list');

            var selection = $A(this._list.getSelectedItemsIds());

            this._type = this._list.getStorageType();
            this._folder = this._list.getCurrentFolder();

            this._selected = $A();

            this._foldersCount = 0;
            this._filesCount = 0;

            this._list.getData().each(function (dataItem) {
                var itemIndex = selection.indexOf(dataItem.id);
                if (itemIndex < 0) {
                    return;
                }
                this._selected.push(dataItem);
                if ('folder' === dataItem.type) {
                    this._foldersCount++;
                } else {
                    this._filesCount++;
                }
            }, this);

            if (0 === this._foldersCount + this._filesCount) {
                view.showWarning(confirmationDialogLocale.itemsNotSelectedOnDelete);
                return;
            }

            var message = '';
            if (this._foldersCount && this._filesCount) {
                message = confirmationDialogLocale.textOnDeleteFilesAndFolders;
            } else if (1 === this._foldersCount) {
                message = confirmationDialogLocale.textOnDeleteFolder;
            } else if (1 === this._filesCount) {
                message = confirmationDialogLocale.textOnDeleteFile;
            } else if (this._foldersCount) {
                message = confirmationDialogLocale.textOnDeleteFolders;
            } else if (this._filesCount) {
                message = confirmationDialogLocale.textOnDeleteFiles;
            }

            var fileName = 1 === this._selected.length
                ? this._selected.first().name
                : '';

            Jsw.messageBox.show({
                type: Jsw.messageBox.TYPE_YESNO,
                text: message
                    .replace('%%count%%', this._foldersCount + this._filesCount)
                    .replace('%%foldersCount%%', this._foldersCount)
                    .replace('%%filesCount%%', this._filesCount)
                    .replace('%%fileName%%', '<b>' + fileName.escapeHTML() + '</b>'),
                subtype: 'delete',
                description: '',
                buttonTitles: {
                    yes: confirmationDialogLocale.buttonYes,
                    no: confirmationDialogLocale.buttonNo,
                },
                onYesClick: this._onConfirmationYesClick.bind(this),
            });
        },

        _onConfirmationYesClick: function () {
            var ids = $H();
            var count = 0;
            this._selected.each(function (dataItem) {
                ids.set('ids[' + count + ']', dataItem.id);
                count++;
            });

            new Ajax.Request(
                Jsw.prepareUrl('/file/delete'),
                {
                    method: 'post',
                    parameters: ids.merge({
                        type: this._type,
                        folder: this._folder,
                    }),
                    onCreate: this._onCreate.bind(this),
                    onSuccess: this._onSuccess.bind(this),
                    onException: this._onException.bind(this),
                }
            );
        },

        _onCreate: function () {
            this._waitDialog = new Wait({
                cls: 'popup-panel',
                title: this._locale.title,
                message: this._locale.waitMessage,
                titleCls: 'pp-delete',
            });
        },

        _onSuccess: function (transport) {
            try {
                var response = transport.responseText.evalJSON();
                switch (response.status) {
                    case 'SUCCESS':
                        this._showSuccessMessage();
                        break;
                    case 'ERROR':
                        if (response.data) {
                            this._onExecuteError(response.message, $H(response.data));
                        } else {
                            this._userError(response.message);
                        }
                        break;
                    default:
                        this._internalError('Unknown response status: ' + response.status);
                }
                Smb.FileSharing.ListInstance.reload();
            } catch (e) {
                this._internalError(transport.responseText);
            }
        },

        _userError: function (message) {
            this.close();

            Jsw.clearStatusMessages();
            Jsw.addStatusMessage('error', message);
        },

        _internalError: function (message) {
            this.close();

            Jsw.clearStatusMessages();
            Jsw.addStatusMessage('error', 'Internal error: ' + message);
        },

        _showSuccessMessage: function () {
            var message = '';
            if (this._foldersCount && this._filesCount) {
                message = this._locale.textOnDeleteFilesAndFolders;
            } else if (1 === this._foldersCount) {
                message = this._locale.textOnDeleteFolder;
            } else if (1 === this._filesCount) {
                message = this._locale.textOnDeleteFile;
            } else if (this._foldersCount) {
                message = this._locale.textOnDeleteFolders;
            } else if (this._filesCount) {
                message = this._locale.textOnDeleteFiles;
            }

            var fileName = 1 === this._selected.length
                ? this._selected.first().name
                : '';

            this.close();

            Jsw.addStatusMessage('info', message
                .replace('%%count%%', this._foldersCount + this._filesCount)
                .replace('%%foldersCount%%', this._foldersCount)
                .replace('%%filesCount%%', this._filesCount)
                .replace('%%fileName%%', '<b>' + fileName.escapeHTML() + '</b>')
            );
        },

        _onExecuteError: function (errorMessage, items) {
            var errorItems = $A();

            this._selected.each((function (listItem) {
                var errorItem = Object.clone(listItem);
                items.each(function (pair) {
                    if (listItem.id == pair.key) {
                        errorItem.status = pair.value;
                        throw $break;
                    }
                }, this);

                errorItems.push(errorItem);
            }).bind(this));

            var message = null;
            if (1 < errorItems.length) {
                message = this._locale.failed;
            }

            this.close();

            this._errorDialog = new ErrorDialog({
                errorMessage: errorMessage,
                message: message,
                locale: this._errorDialogLocale,
                items: errorItems,
            });
        },

        _onException: function (transport, exception) {
            this._internalError(exception);
        },

        close: function () {
            if (this._waitDialog) {
                this._waitDialog.hide();
                this._waitDialog.remove();
                this._waitDialog = null;
            }

            if (this._errorDialog) {
                this._errorDialog.hide();
                this._errorDialog.remove();
                this._errorDialog = null;
            }
        },
    });
});
