// Copyright 1999-2017. Plesk International GmbH. All rights reserved.

define([
    './main',
    './view',
], function (fileSharing, view) {
    return Class.create(Jsw.PopupForm, {

        _initConfiguration: function ($super, config) {
            $super(config);

            this._selectUsersBoxId = this._id + '-select-users';
            this._expiredLinksBlockId = this._id + '-expired-links';
            this._linkTypeSharedId = this._id + '-link-shared';
            this._linkTypePublicId = this._id + '-link-public';
            this._linkExpirationTimeId = this._id + '-link-expiration-time';
            this._additionalAddressesId = this._id + '-additional-email';
            this._emailSubjectId = this._id + '-email-subject';
            this._emailBodyId = this._id + '-email-body';
            this._selectLinkTypeBlockId = this._id + '-link-type-block';

            this._handlerUrl = Jsw.prepareUrl('/file/email');
            this._embedded = this._getConfigParam('embedded', false);

            this._title = this._getConfigParam('title', this.lmsg('title'));

            this._storageType = this._getConfigParam('storageType', '');
            this._folder = this._getConfigParam('folder', '');

            this._onBack = this._getConfigParam('onBack', null);
            this._onOk = this._getConfigParam('onOk', null);
            this._onCancel = this._getConfigParam('onCancel', null);

            this._prependValidation = this._onOk ? true : false;

            this._getFilesHandler = this._getConfigParam('getFilesHandler', null);

            this._selected = $A();
            this._users = $H(Smb.FileSharing.ListInstance.getUsers());
            this._userContactName = Smb.FileSharing.ListInstance.getUserContactName();

            this._displayStatusMessages = this._getConfigParam('displayStatusMessages', false);
        },

        setTargetType: function (type) {
            this._storageType = type;
        },

        setTargetFolder: function (folder) {
            this._folder = folder;
        },

        render: function ($super) {
            $super();

            this.setBoxType('form-box');
            this.setTitle(this._title);
            this.setHint1(this.lmsg('hint'));

            this._renderLinkTypeSelector();
            this._renderSelectUsersDialog();
            this._renderEmailFields();

            this.addBackButton();
            this.addOkButton();
            this.addCancelButton();

            this.reset();
        },

        _getSelectedFiles: function () {
            if (this._getFilesHandler) {
                return this._getFilesHandler();
            }

            var selected = $A();
            var ids = $A();
            Smb.FileSharing.ListInstance.getSelectedItemsIds().each(function (id) {
                ids.push(id);
            });

            this._folderSelected = false;
            Smb.FileSharing.ListInstance.getData().each((function (dataItem) {
                var itemIndex = ids.indexOf(dataItem.id);
                if (itemIndex < 0) {
                    return;
                }

                if ('folder' === dataItem.type) {
                    this._folderSelected = true;
                } else {
                    selected.push(dataItem.name);
                }
            }).bind(this));

            if (selected.length < 1) {
                this.hide();
                view.showWarning(this.lmsg('nothingSelected'));
                return $A();
            }

            if (!this._embedded && this._folderSelected) {
                this.hide();
                view.showWarning(this.lmsg('folderSelected'));
                return $A();
            }

            return selected;
        },

        reset: function () {
            this._selected = this._getSelectedFiles();
            if (!this._selected.length) {
                return false;
            }

            $(this._emailSubjectId).value = this._getDefaultEmailSubject();
            $(this._emailBodyId).value = this._getDefaultEmailBody();

            this._updateLinkTypeSelectionBlock();

            return true;
        },

        _getDefaultEmailSubject: function () {
            return this.lmsg(1 === this._selected.length ? 'emailSubjectOneFile' : 'emailSubjectSeveralFiles', {
                file: this._selected[0].escapeHTML(),
            });
        },

        _getDefaultEmailBody: function () {
            return this.lmsg('emailBodyText', {
                user: this._userContactName,
                links_placeholder: '<- ' + this.lmsg('emailLinksPlaceHolderText') + ' ->',
            });
        },

        _renderUsersList: function () {
            $(this._selectUsersBoxId).update('');
            this._users.each((function (pair) {
                var user = pair.value;
                var inputId = this._id + '-user-block-' + user.id;
                $(this._selectUsersBoxId).insert(
                    '<label for="' + inputId + '">' +
                        '<span>' +
                            '<input id="' + inputId + '" type="checkbox" class="checkbox" value="' + user.id + '" /> ' +
                            user.contactName.escapeHTML() + ' &lt;' + user.email + '&gt;' +
                        '</span>' +
                    '</label>'
                );
            }).bind(this));

            this._correctUsersListElementWidth();

            $(this._selectUsersBoxId).select('input').each(function (element) {
                element.observe('click', this._onSelectUsersItemClick.bindAsEventListener(this));
            }, this);
        },

        _correctUsersListElementWidth: function () {
            // CSS trick
            var maxWidth = 0;
            $(this._selectUsersBoxId).select('span').each(function (element) {
                maxWidth = [Element.getWidth(element), maxWidth].max();
            }, this);

            var innerElementPadding = 12;
            if (maxWidth + innerElementPadding < Element.getWidth($(this._selectUsersBoxId))) {
                return;
            }

            $(this._selectUsersBoxId).select('label').each(function (element) {
                element.setStyle({ width: maxWidth + 'px' });
            }, this);
        },

        _isLinkTypeSelectionBlockRequired: function () {
            return 'private' !== this._storageType;
        },

        _isLinkExpirationBlockRequired: function () {
            return 'protected' !== this._storageType && 'external' !== this._storageType;
        },

        _updateLinkTypeSelectionBlock: function () {
            if (this._isLinkTypeSelectionBlockRequired()) {
                $(this._selectLinkTypeBlockId).update(
                    '<div class="single-row">' +
                        '<div class="field-value">' +
                            '<label>' +
                                '<input type="radio" class="radio" name="linkType" id="' + this._linkTypeSharedId + '" checked="checked" /> ' +
                                this.lmsg('linkTypeShared') +
                            '</label>' +
                        '</div>' +
                        '<div class="field-value">' +
                            '<div class="indent-box">' +
                                '<input type="radio" class="radio" name="linkType" id="' + this._linkTypePublicId + '" />' +
                                '<div class="indent-box-content">' +
                                    '<label for="' + this._linkTypePublicId + '">' + this.lmsg('linkTypePublic') + '</label>' +
                                    '<div id="' + this._expiredLinksBlockId + '" class="expired-info" />' +
                                '</div>' +
                            '</div>' +
                        '</div>' +
                    '</div>'
                );
                $(this._linkTypeSharedId).observe('click', this._onLinkSharedClick.bindAsEventListener(this));
                $(this._linkTypePublicId).observe('click', this._onLinkPublicClick.bindAsEventListener(this));
            } else {
                $(this._selectLinkTypeBlockId).update(
                    '<div class="single-row">' +
                        '<div class="field-value">' +
                            '<label>' + this.lmsg('linkTypePublic') + '</label>' +
                            '<div id="' + this._expiredLinksBlockId + '" class="expired-info" />' +
                        '</div>' +
                    '</div>'
                );
            }

            if (this._isLinkExpirationBlockRequired()) {
                $(this._expiredLinksBlockId).insert(
                    this.lmsg('linkExpiredIn') + ' ' +
                    '<select id="' + this._linkExpirationTimeId + '" disabled="disabled">' +
                        '<option value="24">' + this.lmsg('linkExpiredOneDay') + '</option>' +
                        '<option value="168" selected="selected">' + this.lmsg('linkExpiredOneWeek') + '</option>' +
                        '<option value="720">' + this.lmsg('linkExpiredOneMonth') + '</option>' +
                        '<option value="-1">' + this.lmsg('linkExpiredUnlimited') + '</option>' +
                    '</select>'
                );
            }

            if (!this._isLinkTypeSelectionBlockRequired()) {
                $(this._linkExpirationTimeId).disabled = false;
            }
        },

        _renderLinkTypeSelector: function () {
            $(this._contentAreaId).insert('<div class="form-row" id="' + this._selectLinkTypeBlockId + '" />');
        },

        _onLinkSharedClick: function () {
            if ($(this._linkExpirationTimeId)) {
                $(this._linkExpirationTimeId).disabled = true;
            }
        },

        _onLinkPublicClick: function () {
            if ($(this._linkExpirationTimeId)) {
                $(this._linkExpirationTimeId).disabled = false;
            }
        },

        _renderSelectUsersDialog: function () {
            $(this._contentAreaId).insert(
                '<hr class="hr" />' +
                '<div class="two-cols-block clearfix">' +
                    '<div class="first-col">' +
                        '<div class="col-content">' +
                            '<div class="col-heading">' + this.lmsg('selectUsers') + '</div>' +
                            '<div class="select-block" id="' + this._selectUsersBoxId + '" > </div>' +
                        '</div>' +
                    '</div>' +
                    '<div class="second-col">' +
                        '<div class="col-content">' +
                            '<div class="col-heading">' + this.lmsg('enterAdditionalAddresses') + '</div>' +
                            '<textarea id="' + this._additionalAddressesId + '"></textarea>' +
                            '<div class="hint">' + this.lmsg('enterAdditionalAddressesHint') + '</div>' +
                        '</div>' +
                    '</div>' +
                '</div>'
            );

            this._renderUsersList();
        },

        _onSelectUsersItemClick: function (event) {
            var element = Event.element(event);
            if (element.checked) {
                element.up('label').addClassName('selected');
            } else {
                element.up('label').removeClassName('selected');
            }
        },

        _renderEmailFields: function () {
            $(this._contentAreaId).insert(
                '<div class="form-row">' +
                    '<div class="single-row">' +
                        '<div class="field-name">' + this.lmsg('emailSubject') + '</div>' +
                        '<div class="field-value"><input id="' + this._emailSubjectId + '" type="text" class="input-text f-max-size" /></div>' +
                    '</div>' +
                '</div>' +
                '<div class="form-row">' +
                    '<div class="single-row">' +
                        '<div class="field-name">' + this.lmsg('emailBody') + '</div>' +
                        '<div class="field-value"><textarea id="' + this._emailBodyId + '" class="f-max-size"></textarea></div>' +
                    '</div>' +
                '</div>'
            );
        },

        _onOk: function () {
            this._execute();
        },

        _execute: function () {
            var params = this.getRequestParams();

            new Ajax.Request(
                this._handlerUrl,
                {
                    method: 'post',
                    parameters: params,
                    onCreate: this._onExecuteCreate.bind(this),
                    onSuccess: this._onExecute.bind(this),
                    onException: (function (transport, exception) {
                        this._internalError(exception + '\n' + transport.responseText);
                    }).bind(this),
                    onComplete: this._onExecuteComplete.bind(this),
                }
            );
        },

        getRequestParams: function () {
            var params = $H({
                linkExpirationTime: this._isLinkExpirationBlockRequired() ? $(this._linkExpirationTimeId).value : 0,
                emailSubject: $(this._emailSubjectId).value,
                emailBody: $(this._emailBodyId).value,
                additionalEmailAddresses: $(this._additionalAddressesId).value,
            });

            if (this._isLinkTypeSelectionBlockRequired()) {
                params.set('linkType', $(this._linkTypePublicId).checked ? 'public' : 'shared');
            } else {
                params.set('linkType', 'public');
            }

            params.set('targetType', this._storageType);
            params.set('targetFolder', this._folder);

            var number = 0;
            if (!this._embedded) {
                this._selected.each(function (item) {
                    params.set('items[' + number++ + ']', item);
                });
            }

            this._selectedUsers = $A();
            $(this._selectUsersBoxId).select('input').each(function (element) {
                if (element.checked) {
                    this._selectedUsers.push(element.value);
                }
            }, this);

            number = 0;
            this._selectedUsers.each(function (item) {
                params.set('users[' + number++ + ']', item);
            });

            if (this._prependValidation) {
                params.set('validationOnly', 1);
            }

            return params;
        },

        _onExecute: function (transport) {
            try {
                var result = transport.responseText.evalJSON();
            } catch (exception) {
                this._internalError(exception + '\n' + transport.responseText);
                return;
            }

            switch (result.status) {
                case 'SUCCESS':
                    if (this._prependValidation) {
                        this._prependValidation = false;
                        if (Object.isFunction(this._onOk)) {
                            this._onOk();
                        }
                    } else {
                        this._close();
                        this._showSuccessMessage();
                    }
                    break;
                case 'ERROR':
                    switch (result.data) {
                        case 'noRecipients':
                            this._addErrorMessage(this.lmsg('errorNoEmailAddresses'));
                            break;
                        case 'badRecipients':
                            this._addErrorMessage(this.lmsg('errorWrongEmailAddresses'));
                            $(this._additionalAddressesId).insert({
                                after: '<span class="error-hint">' + this.lmsg('hintWrongEmailAddresses') + '</span>',
                            });
                            break;
                        default:
                            this._addErrorMessage(result.message);
                    }
                    break;
                default:
                    this._internalError(
                        'Unknown response status: ' + result.status
                    );
            }
        },

        _onExecuteCreate: function (transport) {
            this.disable();
        },

        _onExecuteComplete: function (transport) {
            this.enable();
        },

        disable: function () {
            var context = this;
            $(this._rightActionButtonsAreaId).select('.btn.action').each(
                function (actionButtonWrapper) {
                    var sendButton = actionButtonWrapper.down('button');
                    context._okButtonTitle = sendButton.innerHTML;
                    sendButton.update(
                        '<span class="wait">' + context.lmsg('waitButtonTitle') + '</span>'
                    );
                }
            );

            $(this._rightActionButtonsAreaId).select('.btn').each(
                function (actionButtonWrapper) {
                    actionButtonWrapper.down('button').disabled = true;
                    actionButtonWrapper.addClassName('disabled');
                }
            );
        },

        enable: function () {
            var context = this;
            $(this._rightActionButtonsAreaId).select('.btn.action').each(
                function (actionButtonWrapper) {
                    var sendButton = actionButtonWrapper.down('button');
                    sendButton.update(context._okButtonTitle);
                }
            );
            $(this._rightActionButtonsAreaId).select('.btn').each(
                function (actionButtonWrapper) {
                    actionButtonWrapper.down('button').disabled = false;
                    actionButtonWrapper.removeClassName('disabled');
                }
            );
        },

        _getFileUrl: function (file) {
            return fileSharing.getFileUrl(
                Smb.FileSharing.ListInstance.getStorageTypes().get(this._storageType).url,
                this._folder,
                file
            );
        },

        getSuccessMessage: function () {
            if (1 === this._selected.length) {
                return this.lmsg('statusOneFile', {
                    file: '<a href="' + this._getFileUrl(this._selected[0]) + '" target="_blank">' + this._selected[0].escapeHTML() + '</a>',
                });
            }

            return this.lmsg('statusFiles', { number: this._selected.length });
        },

        _showSuccessMessage: function () {
            if (this._displayStatusMessages) {
                view.clearStatusMessages();
                Jsw.addStatusMessage('info', this.getSuccessMessage());
            }
        },

        _close: function () {
            this.hide();
        },

        addBackButton: function () {
            if (Object.isFunction(this._onBack)) {
                this.addLeftButton(
                    this.lmsg('buttonBack'),
                    this._onBack,
                    false,
                    true
                );
            }
        },

        addOkButton: function () {
            this.addRightButton(
                this.lmsg('buttonOk'),
                this._onOkButtonClick.bind(this),
                true,
                true
            );
        },

        _clearMessages: function ($super) {
            $super();
            var textElement = $(this._additionalAddressesId);
            if (!textElement) {
                return;
            }
            $(this._additionalAddressesId).up().select('.error-hint')
                .each(function (element) {
                    element.remove();
                });
        },

        _onOkButtonClick: function () {
            this._clearMessages();

            var params = this.getRequestParams();

            if (
                !params.get('additionalEmailAddresses') &&
                !(params.get('users[0]') && params.get('users[0]').strip())
            ) {
                this._addErrorMessage(this.lmsg('errorNoEmailAddresses'));
                return;
            }

            if (this._prependValidation || !Object.isFunction(this._onOk)) {
                this._execute();
            } else {
                this._onOk();
            }
        },

        addCancelButton: function () {
            this.addRightButton(
                this.lmsg('buttonCancel'),
                this._onCancelButtonClick.bind(this),
                false,
                false
            );
        },

        _onCancelButtonClick: function () {
            if (Object.isFunction(this._onCancel)) {
                this._onCancel();
            } else {
                this.hide();
            }
        },
    });
});
