// Copyright 1999-2017. Plesk International GmbH. All rights reserved.

define([
    './view',
    './explorer',
    './confirmation-dialog',
    './error-dialog',
    './notification',
    './wait',
], function (view, Explorer, ConfirmationDialog, ErrorDialog, Notification, Wait) {
    return Class.create(Jsw.PopupForm, {

        _initConfiguration: function ($super, config) {
            $super(config);

            this._foldersExplorerLocale = this._getConfigParam('foldersExplorerLocale', {});
            this._notificationLocale = this._getConfigParam('notificationLocale', {});
            this._confirmationDialoglocale = this._getConfigParam('confirmationDialogLocale', {});
            this._errorDialoglocale = this._getConfigParam('errorDialogLocale', {});

            this._handlerUrl = this._getConfigParam('handlerUrl', '');

            this._explorerBlockId = this._id + '-explorer-block';
            this._warningBlockId = this._id + '-warning-block';
            this._notificationCheckboxId = this._id + '-notification-checkbox';

            this._selected = $A();
            this._foldersNumber = 0;
            this._filesNumber = 0;
            this._notificationDialog = null;
            this._waitDialog = null;
            this._confirmDialog = null;
            this._errorDialog = null;
            this._explorer = null;
        },

        render: function ($super) {
            $super();

            this.setBoxType('form-box');
            this.setTitle(this.lmsg('title'));

            $(this._contentAreaId).insert(
                this._getContentHtml()
            );

            $(this._notificationCheckboxId).observe(
                'click',
                this._onNotificationClick.bindAsEventListener(this)
            );

            this.addOkButton();
            this.addCancelButton();

            this.reset();
            this._explorer = this._initExplorer();
        },

        _getContentHtml: function () {
            return '' +
                '<div>' +
                    this._getAdditionalHtml() +
                    '<div class="form-row" id="' + this._explorerBlockId + '">' +
                        this._getExplorerTitleHtml() +
                    '</div>' +
                    '<div class="form-row" style="display: none;" id="' + this._warningBlockId + '">' +
                        '<div class="single-row">' +
                            '<div class="hint-attention">' +
                            '</div>' +
                        '</div>' +
                    '</div>' +
                    '<div class="form-row">' +
                        '<div class="single-row">' +
                            '<label for="' + this._notificationCheckboxId + '">' +
                                '<input type="checkbox" class="checkbox" value="1" ' +
                                    'id="' + this._notificationCheckboxId + '"/>' +
                                ' ' + this.lmsg('sendNotification') +
                            '</label>' +
                        '</div>' +
                    '</div>' +
                '</div>';
        },

        _getExplorerTitleHtml: function () {
            return '' +
                '<p>' +
                    this.lmsg('selectDestinationFolder') +
                '</p>';
        },

        _getAdditionalHtml: function () {
            return '';
        },

        _hideWarning: function () {
            $(this._warningBlockId).hide();
        },

        _showWarning: function (message) {
            $(this._warningBlockId).down('.hint-attention').update(message);
            $(this._warningBlockId).show();
        },

        _initExplorer: function () {
            return new Explorer({
                id: this._id + 'Explorer',
                renderTo: this._explorerBlockId,
                storageTabs: this._getStorageTabs(),
                storageTypes: this._getStorageTypes(),
                fileTypesIconsPathTemplate: Jsw.skinUrl + '/icons/16/plesk/file-%%type%%.png',
                locale: this._foldersExplorerLocale,
                onSelectFolder: this._onSelectFolder.bind(this),
                onDeselectFolder: this._onDeselectFolder.bind(this),
            });
        },

        _onSelectFolder: function () {
            this.enable();
        },

        _onDeselectFolder: function () {
            this.disable();
        },

        _getStorageTabs: function () {
            var listControl = Jsw.getComponent('files-list');
            return listControl.getStorageTabs();
        },

        _getStorageTypes: function () {
            var listControl = Jsw.getComponent('files-list');
            return listControl.getStorageTypes();
        },

        _resetExplorer: function () {
            if (this._explorer) {
                this._explorer.reset();
            }
        },

        reset: function () {
            this._selected = $A();
            this._foldersNumber = 0;
            this._filesNumber = 0;
            $(this._notificationCheckboxId).checked = false;

            view.hideWarning();

            this._resetExplorer();

            if (this._notificationDialog) {
                this._notificationDialog.remove();
                this._notificationDialog = null;
            }
            if (this._confirmDialog) {
                this._confirmDialog.remove();
                this._confirmDialog = null;
            }
            if (this._errorDialog) {
                this._errorDialog.remove();
                this._errorDialog = null;
            }
            $(this._rightActionButtonsAreaId).select('.btn.action').each((
                function (buttonWrapper) {
                    var button = buttonWrapper.down('button');
                    button.update(this.lmsg('buttonOk'));
                }
            ).bind(this));

            var listControl = Jsw.getComponent('files-list');

            var ids = $A();
            listControl.getSelectedItemsIds().each(function (id) {
                ids.push(id);
            });
            listControl.getData().each((function (dataItem) {
                var itemIndex = ids.indexOf(dataItem.id);
                if (itemIndex < 0) {
                    return;
                }

                this._selected.push(dataItem);

                if ('folder' === dataItem.type) {
                    this._foldersNumber++;
                } else {
                    this._filesNumber++;
                }
            }).bind(this));

            if (1 > this._selected.length) {
                this._close();
                view.showWarning(this.lmsg('nothingSelected'));

                return false;
            }

            var hint = '';
            if (1 === this._selected.length) {
                var fileName = this._selected[0].name;
                hint = this.lmsg(this._foldersNumber ? 'hintOneFolder' : 'hintOneFile', { file: '<b>' + fileName.escapeHTML() + '</b>' });
            } else if (!this._foldersNumber) {
                hint = this.lmsg('hintFiles', { number: this._filesNumber });
            } else if (!this._filesNumber) {
                hint = this.lmsg('hintFolders', { number: this._foldersNumber });
            } else {
                hint = this.lmsg('hintFilesFolders', { foldersNumber: this._foldersNumber, filesNumber: this._filesNumber });
            }
            this.setHint1(hint);

            if (this._filesNumber) {
                $(this._notificationCheckboxId).enable();
            } else {
                $(this._notificationCheckboxId).disable();
            }

            return true;
        },

        addCancelButton: function () {
            this.addRightButton(
                this.lmsg('buttonCancel'),
                this._close.bind(this),
                false,
                false
            );
        },

        _close: function () {
            if (this._notificationDialog) {
                this._notificationDialog.remove();
                this._notificationDialog = null;
            }
            if (this._confirmDialog) {
                this._confirmDialog.remove();
                this._confirmDialog = null;
            }
            if (this._errorDialog) {
                this._errorDialog.remove();
                this._errorDialog = null;
            }
            if (this._waitDialog) {
                this._waitDialog.hide();
            }
            this.hide();
        },

        addOkButton: function () {
            this.addRightButton(
                this.lmsg('buttonOk'),
                this._onOkClick.bind(this),
                true,
                true
            );
        },

        _onOkClick: function () {
            if ($(this._notificationCheckboxId).checked) {
                this.hide();
                if (this._notificationDialog) {
                    this._notificationDialog.show();
                } else {
                    this._notificationDialog = new Notification({
                        id: this._id + '-notification',
                        renderTo: this.getRenderTarget().id + '-notification',
                        cls: 'popup-panel',
                        locale: this._notificationLocale,
                        titleCls: this._titleCls,
                        title: this._notificationLocale.specifyEmailParameters,
                        onBack: this._onNotificationBackClick.bind(this),
                        onOk: this._onNotificationOkClick.bind(this),
                        onCancel: this._onNotificationCancelClick.bind(this),
                        embedded: true,
                    });
                }
                this._notificationDialog.setTargetType(this._explorer.getType());
                this._notificationDialog.setTargetFolder(this._explorer.getFolder());
                this._notificationDialog.reset();
            } else {
                this._execute(false);
            }
        },

        _onNotificationClick: function (event) {
            var element = Event.element(event);

            $(this._rightActionButtonsAreaId).select('.btn.action button').forEach(function (button) {
                button.update(this.lmsg(element.checked ? 'buttonNext' : 'buttonOk'));
            }.bind(this));
        },

        _onNotificationBackClick: function () {
            this._notificationDialog.hide();
            this.show();
        },

        _onNotificationOkClick: function () {
            this._execute(false);
        },

        _onNotificationCancelClick: function () {
            this._close();
        },

        _execute: function (force) {
            var listControl = Jsw.getComponent('files-list');
            var params = $H({
                type: listControl.getStorageType(),
                folder: listControl.getCurrentFolder(),
                targetType: this._explorer.getType(),
                targetFolder: this._explorer.getFolder(),
                force: force ? 'true' : 'false',
            });

            var number = 0;
            this._selected.each(function (item) {
                params.set(
                    'items[' + number++ + ']',
                    item.id
                );
            });

            if ($(this._notificationCheckboxId).checked && this._notificationDialog) {
                params.set('sendNotification', 'true');
                params = params.merge(this._notificationDialog.getRequestParams());
            }

            new Ajax.Request(
                this._handlerUrl,
                {
                    method: 'post',
                    parameters: params,
                    onCreate: this._showWaitDialog.bind(this),
                    onSuccess: this._onExecute.bind(this),
                    onException: (function (transport, exception) {
                        this._internalError(exception);
                    }).bind(this),
                }
            );
        },

        _onExecute: function (transport) {
            try {
                var result = transport.responseText.evalJSON();
            } catch (exception) {
                this._internalError(
                    exception + '\n' +
                    transport.responseText
                );
                return;
            }

            switch (result.status) {
                case 'SUCCESS':
                    this._showSuccessMessage();
                    this._close();
                    break;
                case 'CONFLICT':
                    this._onExecuteConflict($A(result.data));
                    break;
                case 'ERROR':
                    if (result.data) {
                        this._onExecuteError(result.message, $H(result.data));
                    } else {
                        this._userError(
                            result.message
                        );
                    }
                    break;
                default:
                    this._internalError(
                        'Unknown response status: ' + result.status
                    );
            }
        },

        _onExecuteError: function (errorMessage, items) {
            var errorItems = $A();

            this._selected.each(function (listItem) {
                var errorItem = Object.clone(listItem);
                items.each(function (pair) {
                    if (listItem.id == pair.key) {
                        errorItem.status = pair.value;
                        throw $break;
                    }
                });

                errorItems.push(errorItem);
            });

            var message = null;
            if (1 < errorItems.length) {
                message = this.lmsg('failed');
            }
            this._showErrorDialog(errorMessage, message, errorItems);
        },

        _showErrorDialog: function (errorMessage, message, items) {
            this.hide();
            if (this._notificationDialog) {
                this._notificationDialog.hide();
            }
            if (this._confirmDialog) {
                this._confirmDialog.hide();
            }
            if (this._waitDialog) {
                this._waitDialog.hide();
            }

            this._errorDialog = new ErrorDialog({
                id: this._id + '-error-dialog',
                renderTo: this.getRenderTarget().id + '-error-dialog',
                errorMessage: errorMessage,
                message: message,
                locale: this._errorDialoglocale,
                items: items,
            });
        },

        _showSuccessMessage: function () {
            var message = '';
            var listControl = Jsw.getComponent('files-list');

            if (1 === this._selected.length) {
                var fileName = this._selected.first().name;
                if (this._foldersNumber) {
                    message = this.lmsg('statusOneFolder', {
                        file: '<a href="#" id="' + this._id + '-folder-source-link">' + fileName.escapeHTML() + '</a>',
                    });
                } else {
                    var href = this._explorer.getFileUrl(fileName);
                    message = this.lmsg('statusOneFile', {
                        file: '<a href="' + href + '" target="_blank">' + fileName.escapeHTML() + '</a>',
                    });
                }
            } else if (!this._foldersNumber) {
                message = this.lmsg('statusFiles', { number: this._filesNumber });
            } else if (!this._filesNumber) {
                message = this.lmsg('statusFolders', { number: this._foldersNumber });
            } else {
                message = this.lmsg('statusFilesFolders', { foldersNumber: this._foldersNumber, filesNumber: this._filesNumber });
            }

            message = message.replace(
                '%%destination%%',
                '<a href="#" id="' + this._id + '-success-destination-link">' +
                    this._explorer.getEscapedFolderTitle() +
                '</a>'
            );

            view.clearStatusMessages();
            Jsw.addStatusMessage(
                'info',
                message
            );

            listControl.updateFolderLinkAction(
                $(this._id + '-success-destination-link'),
                this._explorer.getType(),
                this._explorer.getFolder()
            );

            if ($(this._id + '-folder-source-link')) {
                listControl.updateFolderLinkAction(
                    $(this._id + '-folder-source-link'),
                    this._explorer.getType(),
                    this._explorer.makeRelativePath(this._selected.first().name)
                );
            }
        },

        _onExecuteConflict: function (items) {
            var conflictedItems = $A();
            var listControl = Jsw.getComponent('files-list');

            items.each(function (conflictedItem) {
                var listItem = null;
                listControl.getData().each(function (dataItem) {
                    if (dataItem.id == conflictedItem.id) {
                        listItem = dataItem;
                        throw $break;
                    }
                });
                if (!listItem) {
                    return;
                }
                conflictedItems.push({
                    newItem: listItem,
                    existItem: conflictedItem,
                });
            });

            this.hide();
            if (this._notificationDialog) {
                this._notificationDialog.hide();
            }
            if (this._errorDialog) {
                this._errorDialog.hide();
            }
            if (this._waitDialog) {
                this._waitDialog.hide();
            }

            this._confirmDialog = new ConfirmationDialog({
                id: this._id + '-confirmation',
                renderTo: this._id + '-confirmation-dialog-box',
                cls: 'popup-panel',
                locale: this._confirmationDialoglocale,
                destination: this._explorer.getEscapedPathTitle(),
                items: conflictedItems,
                overwriteHandler: this._onConflictOverwrite.bind(this),
                skipHandler: this._onConflictSkip.bind(this, conflictedItems),
                cancelHandler: this._close.bind(this),
            });
        },

        _onConflictOverwrite: function () {
            this._execute(true);
        },

        _onConflictSkip: function (items) {
            var selected = $A();
            this._foldersNumber = 0;
            this._filesNumber = 0;

            this._selected.each((function (dataItem) {
                var listItem = null;
                items.each(function (conflictedItem) {
                    if (dataItem.id == conflictedItem.existItem.id) {
                        listItem = dataItem;
                        throw $break;
                    }
                });
                if (listItem) {
                    return;
                }

                selected.push(dataItem);
                if ('folder' === dataItem.type) {
                    this._foldersNumber++;
                } else {
                    this._filesNumber++;
                }
            }).bind(this));

            this._selected = selected;

            if (1 > this._selected.length) {
                this._close();
                view.showWarning(this.lmsg('nothingChanged'));
                return;
            }

            this._execute(false);
        },

        _showWaitDialog: function () {
            this.hide();
            if (this._notificationDialog) {
                this._notificationDialog.hide();
            }
            if (this._confirmDialog) {
                this._confirmDialog.hide();
            }
            if (this._errorDialog) {
                this._errorDialog.hide();
            }

            if (this._waitDialog) {
                this._waitDialog.show();
                return;
            }
            this._waitDialog = new Wait({
                id: this._id + '-wait',
                renderTo: this.getRenderTarget().id + '-wait',
                cls: 'popup-panel',
                title: this.lmsg('title'),
                message: this.lmsg('waitMessage'),
                titleCls: this._titleCls,
            });
        },

        disable: function () {
            $(this._rightActionButtonsAreaId).select('.btn.action').each(
                function (buttonWrapper) {
                    buttonWrapper.down('button').disabled = true;
                    buttonWrapper.addClassName('disabled');
                }
            );
        },

        enable: function () {
            $(this._rightActionButtonsAreaId).select('.btn.action').each(
                function (buttonWrapper) {
                    buttonWrapper.down('button').disabled = false;
                    buttonWrapper.removeClassName('disabled');
                }
            );
        },
    });
});
