// Copyright 1999-2017. Plesk International GmbH. All rights reserved.

/* global onAfterPhpHandlerChange */

define(function () {
    return Class.create(Jsw.Component, {
        _initConfiguration: function ($super, config) {
            $super(config);
            this._defaults = this._getConfigParam('defaults', {});
            this._elementsPrefix = this._getConfigParam('elementsPrefix', '');
            this._insecureHostingCheckbox = this._getConfigParam('insecureHostingCheckbox');
        },

        _showConflictWarning: function (element, defaultValue, resetHandler) {
            Jsw.messageBox.show({
                type: Jsw.messageBox.TYPE_YESNO,
                subtype: 'confirm',
                text: this.lmsg('sureToApplyConflictOptionTitle'),
                description: this.lmsg('sureToApplyConflictOptionDescription'),
                onNoClick: function () {
                    if (resetHandler) {
                        resetHandler();
                    } else if ('input' === element.tagName.toLowerCase() && 'checkbox' === element.type) {
                        // set inverted value and simulate the selection by click to fire event observers
                        element.checked = !defaultValue;
                        element.click();
                    } else if ('select' === element.tagName.toLowerCase()) {
                        element.value = defaultValue;
                    }
                },
                buttonTitles: {
                    yes: this.lmsg('buttonConflictYes'),
                    no: this.lmsg('buttonConflictNo'),
                },
            });
        },

        _checkShellConflict: function (option, element, callback) {
            var shellEnabledElement = $(this._elementsPrefix + 'shellEnabled');
            var shellEnabled = (shellEnabledElement && shellEnabledElement.checked) || !shellEnabledElement;

            if (shellEnabled && element.value !== '/bin/false' && element.value !== 'Login Disabled') {
                var defaultValue = this._defaults[option];
                callback(element, defaultValue, function () {
                    if ('off' === defaultValue || '' === defaultValue) {
                        if (shellEnabledElement) {
                            shellEnabledElement.checked = false;
                            $(this._elementsPrefix + 'shell').disable();
                        } else {
                            // first value is "disable shell", so it's ok
                            defaultValue = element.options.first().value;
                            element.value = defaultValue;
                        }
                    } else {
                        element.value = defaultValue;
                    }
                });
            }
        },

        _checkConflict: function (element, callback) {
            var option = element.id.split('-').last().underscore();

            if ('input' === element.tagName.toLowerCase() && 'checkbox' === element.type) {
                if (!element.checked && ('1' === this._defaults[option] || 'on' === this._defaults[option])) {
                    callback(element, true);
                } else if (element.checked && ('' === this._defaults[option] || 'off' === this._defaults[option])) {
                    callback(element, false);
                }
            } else if ('select' === element.tagName.toLowerCase()) {
                if (option.startsWith('php_handler_id')) {
                    option = 'php_handler_id';
                }
                if ('any' !== this._defaults[option] && element.value !== this._defaults[option]) {
                    if (option === 'shell') {
                        this._checkShellConflict(option, element, callback);
                    } else if ('php_handler' === option) {
                        var defaultValue = this._defaults[option];
                        callback(element, defaultValue, function () {
                            element.value = defaultValue;
                            if (onAfterPhpHandlerChange) {
                                onAfterPhpHandlerChange();
                            }
                        });
                    } else {
                        callback(element, this._defaults[option]);
                    }
                }
            }
        },

        _initComponentElement: function ($super) {
            $super();

            this._elements = [];

            Object.keys(this._defaults).each(function (option) {
                var element = $(this._elementsPrefix + option.dasherize().camelize());
                if (!element) {
                    if ('ssl' === option) {
                        element = $('sslSettings-ssl');
                    } else if ('write_modify' === option) {
                        element = $(this._elementsPrefix + 'write_modify');
                    } else if ('iis_app_pool' === option) {
                        element = $('tabs-performanceTab-performanceSubForm-iis_app_pool');
                    } else if ('php_handler' === option) {
                        element = $('tabs-phpSettingsTab-phpSettings-phpHandler');
                    } else if ('php_handler_id' === option) {
                        var handlerHash = this._defaults.php_handler;
                        element = $(this._elementsPrefix + 'phpHandlerId' + handlerHash);
                        if (!element) {
                            element = $('tabs-phpSettingsTab-phpSettings-phpHandlerId' + handlerHash);
                        }
                    } else if ('php_safe_mode' === option) {
                        element = $('tabs-phpTab-php-general-safe_mode_predefined');
                        // hosting panel
                        if (!element) {
                            element = $('tabs-phpSettingsTab-phpsection-general-safe_mode_predefined');
                        }
                        this._defaults.safe_mode_predefined = this._defaults[option];
                    }
                }

                if (!element) {
                    return;
                }

                this._elements.push(element);
            }.bind(this));

            this._elements.each(function (element) {
                var eventType = ('select' === element.tagName.toLowerCase()) ? 'change' : 'click';
                element.observe(eventType, function () {
                    if (this._insecureHostingCheckbox && this._insecureHostingCheckbox.checked) {
                        return;
                    }

                    this._checkConflict(element, this._showConflictWarning.bind(this));
                }.bind(this));
            }.bind(this));

            var shellEnabledElement = $(this._elementsPrefix + 'shellEnabled');
            var shellElement = $(this._elementsPrefix + 'shell');
            if (shellEnabledElement) {
                shellEnabledElement.observe('click', function () {
                    if (this._insecureHostingCheckbox && this._insecureHostingCheckbox.checked) {
                        return;
                    }

                    if (shellEnabledElement.checked && ('' === this._defaults.shell || 'off' === this._defaults.shell)) {
                        this._showConflictWarning(shellEnabledElement, false, function () {
                            shellEnabledElement.checked = false;
                            shellElement.disable();
                        });
                    }
                }.bind(this));
            }

            this._componentElement.applySecureSettings = this._applySecureSettings.bind(this);
        },

        _addEvents: function () {
            this._componentElement.submit = this._componentElement.submit.wrap(this._onSubmit.bind(this));
            if (this._insecureHostingCheckbox) {
                this._insecureHostingCheckbox.observe('click', function () {
                    if (this._insecureHostingCheckbox.checked) {
                        return;
                    }

                    this._applySecureSettings();
                }.bind(this));
            }
        },

        _onSubmit: function (origSubmitHandler) {
            if ((this._insecureHostingCheckbox && this._insecureHostingCheckbox.checked)
                || ($(this._elementsPrefix + 'hostingEnabled') && !$(this._elementsPrefix + 'hostingEnabled').checked)
            ) {
                return origSubmitHandler();
            }

            var conflictsFound = false;
            var conflictOptionsHtml = '<ul class="ul">';

            this._elements.each(function (element) {
                this._checkConflict(element, function () {
                    var option = element.id.split('-').last().underscore();
                    if (option.startsWith('php_handler_id')) {
                        option = 'php_handler_id';
                    }
                    conflictOptionsHtml += '<li>' + this.lmsg('option_' + option) + '</li>';
                    conflictsFound = true;
                }.bind(this));
            }.bind(this));

            conflictOptionsHtml += '</ul>';

            if (!conflictsFound) {
                return origSubmitHandler();
            }

            Jsw.messageBox.show({
                type: Jsw.messageBox.TYPE_YESNO,
                subtype: 'confirm',
                text: this.lmsg('foundConflictOptionsTitle'),
                description: this.lmsg('foundConflictOptionsDescription') + conflictOptionsHtml,
                descriptionWrapperTag: 'div',
                onYesClick: function () {
                    origSubmitHandler();
                },
                buttonTitles: {
                    yes: this.lmsg('buttonConflictYes'),
                    no: this.lmsg('buttonConflictNo'),
                },
            });

            return false;
        },

        _applySecureSettings: function () {
            this._elements.each(function (element) {
                var option = element.id.split('-').last().underscore();
                if ('shell' === option && ('' === this._defaults[option] || 'off' === this._defaults[option])) {
                    if ($(this._elementsPrefix + 'shellEnabled')) {
                        $(this._elementsPrefix + 'shellEnabled').checked = false;
                        $(this._elementsPrefix + 'shell').disable();
                    }
                } else if ('php_handler' === option && 'any' !== this._defaults[option]) {
                    element.value = this._defaults[option];
                    if (onAfterPhpHandlerChange) {
                        onAfterPhpHandlerChange();
                    }
                } else if (option.startsWith('php_handler_id') && 'any' !== this._defaults[option]) {
                    element.value = this._defaults.php_handler_id;
                } else if ('input' === element.tagName.toLowerCase() && 'checkbox' === element.type) {
                    if ('1' === this._defaults[option] || 'on' === this._defaults[option]) {
                        element.checked = true;
                    } else if ('' === this._defaults[option] || 'off' === this._defaults[option]) {
                        element.checked = false;
                    }

                    // invert the value and simulate the selection by click to fire event observers
                    element.checked = !element.checked;
                    element.click();
                } else if ('select' === element.tagName.toLowerCase() && 'any' !== this._defaults[option]) {
                    element.value = this._defaults[option];
                }
            }.bind(this));
        },
    });
});
