// Copyright 1999-2017. Plesk International GmbH. All rights reserved.

define(function () {
    return Class.create(Jsw.List, {
        _initConfiguration: function ($super, config) {
            config = Object.extend({
                id: 'log-file-list',
                searchOveral: 'name',
                isDisabledItem: function (item) {
                    return item.readOnly;
                },
            }, config || {});
            $super(config);

            this._webspaceId = this._getConfigParam('webspaceId', '');
            this._urls = this._getConfigParam('urls', {});

            this._initActions();
            this._initColumns();
            this._initOperations();
        },

        _initColumns: function () {
            this._columns = [
                Jsw.list.COLUMN_SELECTION,
                {
                    header: this.lmsg('modificationDate'),
                    sortable: true,
                    dataIndex: 'modificationDate',
                    cls: 't-r number',
                    renderer: function (item) {
                        return item.modificationDisplayDate.escapeHTML();
                    },
                },
                {
                    header: this.lmsg('name'),
                    sortable: true,
                    dataIndex: 'name',
                    renderer: function (item) {
                        return '<span class="tooltipData">' + this.lmsg('viewHint') + '</span>' +
                            '<a href="' +
                            this._urls.browser + '?file=' + encodeURIComponent(item.id) +
                            '" class="log-file-name">' + item.name.escapeHTML() + '</a>';
                    }.bind(this),
                },
                {
                    header: this.lmsg('path'),
                    sortable: true,
                    dataIndex: 'path',
                },
                {
                    header: this.lmsg('fileSize'),
                    sortable: true,
                    dataIndex: 'fileSize',
                    cls: 't-r number',
                    renderer: function (item) {
                        return item.fileDisplaySize.escapeHTML();
                    },
                },
                {
                    header: '',
                    sortable: false,
                    cls: 'action-icon-set t-r min',
                    renderer: this._renderTools.bind(this),
                },
            ];
        },

        _renderTools: function (item) {
            var buttons = new Element('span');
            if (item.isCustom) {
                var unlinkTitle = this.lmsg('buttonUnlinkHint');
                var unlinkTooltip = new Element('span', { class: 'tooltipData' }).insert(unlinkTitle);
                var unlinkButton = new Element('a', {
                    class: 'i-link',
                    href: '#',
                    'data-action-name': 'unlinkCustomLog',
                })
                    .insert(new Element('i', { class: 'icon-unlink' }));
                buttons.insert(new Element('span').insert(unlinkTooltip).insert(unlinkButton));
            }

            var viewTitle = this.lmsg('buttonViewHint');
            var viewTooltip = new Element('span', { class: 'tooltipData' }).insert(viewTitle);
            var viewButton = new Element('a', {
                class: 'i-link',
                href: '#',
                'data-action-name': 'viewLog',
            })
                .insert(new Element('i', { class: 'icon-logs' }));
            buttons.insert(new Element('span').insert(viewTooltip).insert(viewButton));

            var downloadUrl = this._urls.download + '?filename=' + encodeURIComponent(item.id);
            var downloadTitle = this.lmsg('buttonDownloadHint');
            var downloadTooltip = new Element('span', { class: 'tooltipData' }).insert(downloadTitle);
            var downloadButton = new Element('a', { href: downloadUrl, class: 'i-link' })
                .insert(new Element('i', { class: 'icon-download' }));
            buttons.insert(new Element('span').insert(downloadTooltip).insert(downloadButton));

            return buttons.innerHTML;
        },

        _initOperations: function () {
            this._operations = [];
            this._operations.push({
                componentType: 'Jsw.SmallButton',
                id: 'buttonAddCustomLog',
                title: this.lmsg('addCustomLog'),
                description: this.lmsg('addCustomLogHint'),
                addCls: 'sb-add-new',
                handler: this._addCustomLog.bind(this),
            });

            if (this._urls.logRotation) {
                this._operations.push({
                    componentType: 'Jsw.SmallButton',
                    id: 'buttonSettings',
                    title: this.lmsg('logRotation'),
                    description: this.lmsg('logRotationHint'),
                    addCls: 'sb-log-rotation',
                    handler: function () {
                        new Jsw.AjaxPopupForm({
                            cls: 'popup-panel',
                            title: this.lmsg('logRotationTitle'),
                            url: this._urls.logRotation,
                            locale: {
                                waitMsg: this.lmsg('loadingTitle'),
                            },
                        });
                    }.bind(this),
                });
                this._operations.push({
                    componentType: 'Jsw.bar.Separator',
                });
            }

            this._operations.push({
                componentType: 'Jsw.SmallButton',
                id: 'buttonRemoveLogFiles',
                title: this.lmsg('remove'),
                description: this.lmsg('hintRemove'),
                addCls: 'sb-remove-selected',
                handler: function (event) {
                    this.execGroupOperation({
                        url: this._urls.deleteFile,
                        subtype: 'delete',
                        mouseEvent: event,
                        locale: {
                            confirmOnGroupOperation: this.lmsg('confirmOnDelete'),
                        },
                    });
                }.bind(this),
            });
        },

        _addCustomLog: function () {
            new Jsw.DirectoryBrowser({
                subscriptionId: this._webspaceId,
                showFiles: true,
                treeConfig: {
                    filterNodes: function (item) {
                        return item.isDirectory || 'text' === item.type || 'unknown_text' === item.type;
                    },
                },
                locale: {
                    title: this.lmsg('customLogPopupTitle'),
                    buttonOk: this.lmsg('customLogPopupOk'),
                    buttonCancel: this.lmsg('customLogPopupCancel'),
                },
                onSubmit: function (path) {
                    new Ajax.Request(this._urls.addCustom, {
                        method: 'post',
                        parameters: { file: '/' + path },
                        onCreate: function () {
                            this.disable();
                        }.bind(this),
                        onComplete: function (transport) {
                            this.reload();

                            var response = transport.responseText.evalJSON();
                            if (response.isUnsupportedLog) {
                                Jsw.addStatusMessage('warning', this.lmsg('invalidLogFormat'));
                            }
                        }.bind(this),
                    });
                }.bind(this),
            });
        },

        _unlinkCustomLog: function (item) {
            Jsw.messageBox.show({
                subtype: 'unlink',
                text: this.lmsg('buttonUnlinkConfirmationDescription'),
                type: Jsw.messageBox.TYPE_YESNO,
                buttonTitles: {
                    yes: this.lmsg('buttonUnlinkConfirmationYes'),
                    no: this.lmsg('buttonUnlinkConfirmationNo'),
                },
                onYesClick: function () {
                    new Ajax.Request(this._urls.unlink, {
                        method: 'post',
                        parameters: { file: item.id },
                        onCreate: function () {
                            this.disable();
                        }.bind(this),
                        onComplete: function () {
                            this.reload();
                        }.bind(this),
                    });
                }.bind(this),
            });
        },

        _initActions: function () {
            this._itemActions.unlinkCustomLog = this._unlinkCustomLog.bind(this);

            this._itemActions.viewLog = function (item) {
                var viewUrl = this._urls.view + '?filename=' + encodeURIComponent(item.id);
                window.open(viewUrl, '_blank', 'toolbar=yes,scrollbars=yes,resizable=yes,width=600,height=500,innerWidth=1000,innerHeight=500');
            }.bind(this);
        },
    });
});
