// Copyright 1999-2017. Plesk International GmbH. All rights reserved.

define([
    './popup-package-info',
    './popup-packages-lock',
    './popup-packages-unlock',
    './popup-packages-update',
], function (PopupPackageInfo, PopupPackagesLock, PopupPackagesUnlock, PopupPackagesUpdate) {
    return Class.create(Jsw.List, {
        _initConfiguration: function ($super, config) {
            config = Object.extend({
                searchOveral: 'name',
                onRedraw: function () {
                    this._renderAlert();
                }.bind(this),
            }, config || {});
            $super(config);

            this._urls = this._getConfigParam('urls', {});

            this._initActions();
            this._initOperations();
            this._initColumns();
        },

        _initActions: function () {
            this._itemActions.packageInfo = function (item) {
                new PopupPackageInfo({
                    locale: this.getLocale(),
                    url: this._urls.packageInfo,
                    packageName: item.name,
                });
            }.bind(this);
            this._itemActions.packageUpdate = function (item) {
                new PopupPackagesUpdate({
                    locale: this.getLocale(),
                    urls: this._urls,
                    packages: [item],
                });
            }.bind(this);
            this._itemActions.packageLock = function (item) {
                new PopupPackagesLock({
                    locale: this.getLocale(),
                    urls: this._urls,
                    packages: [item],
                    list: this,
                });
            }.bind(this);
            this._itemActions.packageUnlock = function (item) {
                new PopupPackagesUnlock({
                    locale: this.getLocale(),
                    urls: this._urls,
                    packages: [item],
                    list: this,
                });
            }.bind(this);
        },

        _initOperations: function () {
            this._operations = [];
            if (this._urls.updatePackages) {
                this._operations.push({
                    componentType: 'Jsw.SmallButton',
                    id: 'buttonUpdate',
                    title: this.lmsg('buttonUpdate'),
                    description: this.lmsg('buttonUpdateDescription'),
                    addCls: 'sb-upgrade',
                    handler: function (event) {
                        this.execGroupOperation({
                            url: this._urls.updatePackages,
                            mouseEvent: event,
                            confirmationPopup: function () {
                                new PopupPackagesUpdate({
                                    locale: this.getLocale(),
                                    urls: this._urls,
                                    packages: this.getSelectedItems(),
                                });
                            }.bind(this),
                        });
                    }.bind(this),
                });
                this._operations.push({
                    componentType: 'Jsw.bar.Separator',
                });
            }
            if (this._urls.lock && this._urls.unlock) {
                this._operations.push({
                    componentType: 'Jsw.SmallButton',
                    id: 'buttonLock',
                    title: this.lmsg('buttonLock'),
                    description: this.lmsg('buttonLockDescription'),
                    addCls: 'sb-lock',
                    handler: function (event) {
                        this.execGroupOperation({
                            url: this._urls.lock,
                            mouseEvent: event,
                            confirmationPopup: function () {
                                new PopupPackagesLock({
                                    locale: this.getLocale(),
                                    urls: this._urls,
                                    packages: this.getSelectedItems(),
                                    list: this,
                                });
                            }.bind(this),
                        });
                    }.bind(this),
                });
                this._operations.push({
                    componentType: 'Jsw.SmallButton',
                    id: 'buttonUnlock',
                    title: this.lmsg('buttonUnlock'),
                    description: this.lmsg('buttonUnlockDescription'),
                    addCls: 'sb-unlock',
                    handler: function (event) {
                        this.execGroupOperation({
                            url: this._urls.unlock,
                            mouseEvent: event,
                            confirmationPopup: function () {
                                new PopupPackagesUnlock({
                                    locale: this.getLocale(),
                                    urls: this._urls,
                                    packages: this.getSelectedItems(),
                                    list: this,
                                });
                            }.bind(this),
                        });
                    }.bind(this),
                });
                this._operations.push({
                    componentType: 'Jsw.bar.Separator',
                });
            }
            this._operations.push({
                componentType: 'Jsw.SmallButton',
                id: 'buttonCheckUpdates',
                title: this.lmsg('buttonCheckUpdates'),
                description: this.lmsg('buttonCheckUpdatesDescription'),
                addCls: 'sb-refresh',
                handler: this.checkUpdates.bind(this),
            });
        },

        _initColumns: function () {
            this._columns.push(Jsw.list.COLUMN_SELECTION);
            this._columns.push({
                header: this.lmsg('name'),
                sortable: true,
                dataIndex: 'name',
                renderer: function (item) {
                    return '<a data-action-name="packageInfo" href="#">'
                        + item.name.escapeHTML() + ' ' + item.version_from.escapeHTML()
                        + ' (' + (item.repo_from ? item.repo_from.escapeHTML() : this.lmsg('unknownRepo')) + ')</a>';
                }.bind(this),
            });
            this._columns.push({
                header: this.lmsg('summary'),
                sortable: true,
                dataIndex: 'summary',
            });
            this._columns.push({
                header: this.lmsg('lockedCut'),
                sortable: true,
                dataIndex: 'locked',
                cls: 'min',
                renderer: function (item) {
                    if (item.locked) {
                        return '<span class="tooltipData">' + this.lmsg('lockedHint') + '</span>' +
                            '<a data-action-name="packageUnlock" href="#"><i class="icon-lock"></i></a>';
                    }
                    return '<span class="tooltipData">' + this.lmsg('unlockedHint') + '</span>' +
                        '<a data-action-name="packageLock" href="#"><i class="icon-unlock-muted"></i></a>';
                }.bind(this),
            });
            this._columns.push({
                header: this.lmsg('updates'),
                sortable: true,
                dataIndex: 'update',
                renderer: function (item) {
                    if (!item.update) {
                        return this.lmsg('updated');
                    }
                    var updateToVersion = this.lmsg('update', { version: item.version_to.escapeHTML() });
                    return '<a data-action-name="packageUpdate" href="#">' + updateToVersion + ' (' + item.repo_to.escapeHTML() + ')</a>';
                }.bind(this),
            });
        },

        _getEmptyHtml: function ($super) {
            if (this._additionalData.updatesOnly && this._additionalData.emptyList) {
                return '' +
                    '<div class="js-empty-list">' +
                    '<div class="extended-status-message-container">' +
                    '<div class="extended-status-message extended-status-message-success">' +
                    '<div class="extended-status-message-image"><img src="' + Jsw.skinUrl + '/images/box.png" alt=""></div>' +
                    '<div class="extended-status-message-body">' +
                    '<div class="extended-status-message-title">' + this.lmsg('allUpToDate') + '</div>' +
                    (this._additionalData.lastChecked ?
                        '<div class="extended-status-message-description">' +
                        this.lmsg('allUpToDateLastChecked', { date: this._additionalData.lastChecked.escapeHTML() }) +
                        '</div>' : '') +
                    '<div class="extended-status-message-actions">' +
                    '<a class="btn" href="#" onclick="Jsw.getComponent(\'' + this._id + '\').checkUpdates(event)">' +
                    '<i class="icon-recheck"></i> ' +
                    this.lmsg('recheckUpdates') +
                    '</a>' +
                    '</div>' +
                    '</div>' +
                    '</div>' +
                    '</div>' +
                    '</div>';
            }
            return $super();
        },

        _renderAlert: function () {
            if (!this._additionalData.availableUpdates.length) {
                return;
            }

            this._componentElement.parentNode.select('.alert').invoke('remove');

            Jsw.render(
                this._componentElement,
                Jsw.createElement('.alert.alert-info', { role: 'alert' }, [
                    Jsw.createElement('i.icon-update-light'),
                    '&nbsp; ',
                    this.lmsg('updatesAvailable', { count: this._additionalData.availableUpdates.length }),
                    this._additionalData.lastChecked ? Jsw.createElement('span.text-muted',
                        ' ' + this.lmsg('lastedChecked', { date: this._additionalData.lastChecked.escapeHTML() })
                    ) : '',
                    '&nbsp; ',
                    Jsw.createElement('a.btn.btn-primary', {
                        onclick: function (event) {
                            event.preventDefault();
                            new PopupPackagesUpdate({
                                locale: this.getLocale(),
                                urls: this._urls,
                                packages: this._additionalData.availableUpdates,
                            });
                        }.bind(this),
                    }, this.lmsg('updateAll')),
                ]),
                'before'
            );
        },

        checkUpdates: function (event) {
            event && event.preventDefault();
            this.reload('?check-updates=true');
        },
    });
});
