// Copyright 1999-2017. Plesk International GmbH. All rights reserved.

Jsw.namespace('Jsw.Task.StatusMessage');

Jsw.Task.StatusMessage = Class.create(Jsw.Component, {

    // see common/php/plib/Db/Table/Row/LongTask.php
    STATUS_NOT_STARTED: 'not_started',
    STATUS_STARTED: 'started',
    STATUS_DONE: 'done',
    STATUS_CANCELED: 'canceled',
    STATUS_ERROR: 'error',

    _failedStatusUpdates: 0,
    _maxFailedStatusUpdates: 10,

    _initConfiguration: function ($super, config) {
        $super(config);

        this._id = this._getConfigParam('id');
        this._onStatusUpdate = this._getConfigParam('onStatusUpdate', null);
        this._onTaskDone = this._getConfigParam('onDone', null);
        this._onTaskCancel = this._getConfigParam('onCancel', null);
        this._onTaskError = this._getConfigParam('onError', null);
        var self = this;
        this._renderMessage = this._getConfigParam('renderMessage', function (status) {
            if (!status.message) {
                return 'Internal error.';
            }
            if (status.progressParams.errorMessage) {
                var messages = status.progressParams.errorMessage.split('\n');
                if (messages.length <= 1) {
                    return status.progressParams.errorMessage;
                }
                var message = messages[0] + ' <a href="#" class="js-details-link">' + self.lmsg('details') + '</a><br>';
                message += '<span class="js-details hidden">';
                for (var i = 1; i < messages.length; i++) {
                    message += messages[i] + '<br>';
                }
                message += '</span>';
                return message;
            }
            return status.message;
        });

        this._statusUrl = this._getConfigParam('statusUrl', Jsw.prepareUrl('/task/status/taskId/'));
        this._updateInterval = this._getConfigParam('updateInterval', 60);
        this._autoRender = true;
    },

    _showProgress: function (status) {
        var statusClass = 'msg-progress';
        if (this.STATUS_DONE === status.status) {
            if (status.progressParams.errorMessage) {
                statusClass = 'msg-warning';
            } else {
                statusClass = 'msg-info';
            }
        } else if (this.STATUS_CANCELED === status.status) {
            statusClass = 'msg-warning';
        } else if (this.STATUS_ERROR === status.status) {
            statusClass = 'msg-error';
        }

        var message = this._renderMessage(status);

        if (this.STATUS_DONE !== status.status && status.progress) {
            if (this._getConfigParam('percent', true)) {
                message = message + ': <span id="backupProgress">' + status.progress + '</span>%';
            }

            if (this._getConfigParam('gauge', true)) {
                message = '<table class="msg-progress-container" cellspacing="0" width="100%"><tr>' +
                        '<td class="msg-progress-info">' + message + '</td>' +
                        '<td class="msg-progress-indicator">' +
                            '<div class="progress progress-sm">' +
                                '<div class="progress-bar" id="backupProgressGauge" style="width: ' + status.progress + '%;"></div>' +
                            '</div>' +
                        '</td>' +
                    '</tr></table>';
            }
        }

        var content = '<div class="msg-box ' + statusClass + '">' +
            '<div class="msg-content">' + message + '</div>' +
            '</div>';

        this._componentElement.update(content);
    },

    _addEvents: function ($super) {
        $super();
        this._processStatus(this._getConfigParam('status', null));
    },

    _scheduleUpdateStatus: function () {
        this._updateStatus.bind(this).delay(this._updateInterval);
    },

    _updateStatus: function () {
        new Ajax.Request(this._statusUrl + this._id, {
            method: 'get',
            onSuccess: this._onStatusUpdateSuccess.bind(this),
            onFailure: this._onStatusUpdateFailure.bind(this),
        });
    },

    _onStatusUpdateSuccess: function (transport) {
        var status = transport.responseText.evalJSON();
        this._failedStatusUpdates = 0;
        this._processStatus(status);
        if (this._onStatusUpdate) {
            this._onStatusUpdate(status);
        }
    },

    _onStatusUpdateFailure: function (transport) {
        this._failedStatusUpdates++;
        if (this._failedStatusUpdates < this._maxFailedStatusUpdates) {
            this._scheduleUpdateStatus();
        }
    },

    _processStatus: function (status) {
        if (!status) {
            this._showLoading();
            this._updateStatus();
        } else if (this.STATUS_DONE === status.status) {
            if (this._onTaskDone) {
                this._onTaskDone(status);
            }
            this._showFinished(status);
        } else if (this.STATUS_CANCELED === status.status) {
            if (this._onTaskCancel) {
                this._onTaskCancel(status);
            }
            this._showFinished(status);
        } else if (this.STATUS_ERROR === status.status) {
            if (this._onTaskError) {
                this._onTaskError(status);
            }
            this._showError(status);
        } else if (this.STATUS_STARTED === status.status || this.STATUS_NOT_STARTED === status.status) {
            this._showWorking(status);
            this._scheduleUpdateStatus();
        } else {
            this._showNone(status);
        }
    },

    _addDetailsEvent: function () {
        var details = this._componentElement.down('.js-details-link');
        if (!details) {
            return;
        }
        var self = this;
        details.observe('click', function (event) {
            event.preventDefault();
            self._componentElement.down('.js-details').toggleClassName('hidden');
            details.toggleClassName('hidden');
        });
    },

    _showFinished: function (status) {
        this._showProgress(status);
        var self = this;
        setTimeout(function () {
            self._addDetailsEvent();
        }, 0);
    },

    _showWorking: function (status) {
        this._showProgress(status);
    },

    _showError: function (status) {
        this._showProgress(status);
    },

    _showLoading: function () {
    },

    _showNone: function (status) {
    },
});
