# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
from __future__ import generators

import os, mimetools
import osutil
from ReadLineAdapter import FileAdapter
import split_files
import gzip_stream
import dump_format
import MultipartParser

import binascii

#
# Errors during backup operations
#
class BackupFailedException(Exception):
    pass

class Metadata(object):
    def __init__(self, fh):
        self.fh = fh

    def getInfo(self):
        if not self.fh:
            raise RuntimeError("Trying to ask invalidated Metadata")
        info = dump_format.readInfo(self.fh)
        self.invalidate()
        return info

    def getDescription(self):
        if not self.fh:
            raise RuntimeError("Trying to ask invalidated Metadata")
        description = dump_format.readDescription(self.fh)
        self.invalidate()
        return description

    def process(self, processor):
        if not self.fh:
            raise RuntimeError("Trying to get invalidated Metadata")
        ret = processor(self.fh)
        self.invalidate()
        return ret

    def extract(self, fileName):
        if not self.fh:
            raise RuntimeError("Trying to ask invalidated Metadata")
        osutil.saveToFile(fileName, self.fh)
        self.invalidate()

    def invalidate(self):
        self.fh = None


class Bundle(object):
    """Bundle is the part of backup carrying content.

    Bundles are generated from MIME-messages, and fluent by default:
    that is, they might be read only one-by-one and stored bundle will
    be invalidated as soon as next bundle is generated.

    Bundle may be fixated: bundle become permanent and it's lifetime
    won't be coupled with source message anymore. This operation
    stores the data in the temporary file, so it is costly and should
    be avoided if possible."""

    class Volatile(object):
        def __init__(self, fh, session_dir):
            self.fh = fh
            self.session_dir = session_dir
        def fixate(self, parent, partId):
            outfile = os.path.join(self.session_dir, "archives/", partId)
            osutil.saveToFile(outfile, self.fh)
            parent._impl = Bundle.Fixated(outfile, True)
        def process(self, parent, processor):
            ret = processor(self.fh)
            parent._impl = Bundle.Consumed()
            return ret
        def invalidate(self, parent):
            parent._impl = Bundle.Invalidated()

    class Fixated(object):
        def __init__(self, filename, temporaryFile = False):
            self.filename = filename
            self.temporaryFile = temporaryFile
        def fixate(self, parent):
            raise RuntimeError("Trying to fixate fixated bundle")
        def process(self, parent, processor):
            fh = file(self.filename, "rb")
            ret = processor(fh)
            fh.close()
            parent._impl = Bundle.Consumed()
            if self.temporaryFile:
                os.unlink(self.filename)
            return ret
        def invalidate(self, parent):
            pass

    class Consumed(object):
        def fixate(self, parent):
            raise RuntimeError("Trying to fixate consumed bundle")
        def process(self, parent, processor):
            raise RuntimeError("Trying to consume consumed bundle")
        def invalidate(self, parent):
            parent._impl = Bundle.Invalidated()

    class Invalidated(object):
        def fixate(self, parent):
            raise RuntimeError("Trying to fixate invalidated bundle")
        def process(self, parent, processor):
            raise RuntimeError("Trying to consume invalidated bundle")
        def invalidate(self, parent):
            raise RuntimeError("Trying to invalidate invalidated bundle")

    def __init__(self, filename = None, fh = None, session_dir = None):
        # fh is volatile, filename is fixated
        if filename:
            self._impl = Bundle.Fixated(filename)
        elif fh:
            self._impl = Bundle.Volatile(fh, session_dir)
        else:
            raise RuntimeError("No filename or fh passed")

    def fixate(self, partId):
        self._impl.fixate(self, partId)

    def process(self, processor):
        return self._impl.process(self, processor)

    def invalidate(self):
        return self._impl.invalidate(self)


#
# Returns the list of pairs (part_info, part_content) processing given fp.
# part_info is the mimetools.Message
# part_content is raw MIME-part data
#
def enumerateParts(fp):
    return enumerateSignedParts(fp, False)

def enumerateSignedParts(fp, test_sign):
    toplevel = mimetools.Message(fp)
    signed = False
    
    if toplevel.gettype() == "multipart/signed":
        boundary = "--" + toplevel.getparam("boundary") + "\n"
        while True:
            line = fp.readline()
            if line == boundary:
                break
        toplevel = mimetools.Message(fp)
        signed = True

    if toplevel.gettype() != "multipart/related":
        raise RuntimeError("Top level is not multipart/related", toplevel.gettype())
    boundary = toplevel.getparam("boundary")

    multipart = MultipartParser.MultipartParser(fp, boundary)
    part_content = FileAdapter(multipart)
    part_info = mimetools.Message(part_content)
    if test_sign:
        yield (part_info, part_content, signed)
    else:
        yield (part_info, part_content)
    while multipart.next():
        part_content = FileAdapter(multipart)
        part_info = mimetools.Message(part_content)
        if test_sign:
            yield (part_info, part_content, signed)
        else:
            yield (part_info, part_content)


#
# File-like transparent base64 decoder object
# Supplied fh should support .readline() and .close()
#
class Base64Decoder(object):
    def __init__(self, fh):
        self.fh = fh
        self.buffer = ""

    def read(self, size):
        while self.fh and len(self.buffer) < size:
            line = self.fh.readline()
            if not line:
                self.fh.close()
                self.fh = None
                break
            self.buffer += binascii.a2b_base64(line)

        part = self.buffer[:size]
        self.buffer = self.buffer[size:]
        return part

    def close(self):
        if self.fh:
            self.fh.close()
        self.fh = None

def decodePart(encoding, fh):
    if encoding == "binary" or encoding == "8bit" or encoding == "7bit":
        return fh
    elif encoding == "base64":
        return Base64Decoder(fh)
    else:
        raise RuntimeError(encoding + " encoding is not supported")

def decodeParts(parts):
    """Decodes bodies of parts to the plain data. Required for processing
       early 8.0.0 backups, which have base64-encoded parts.
    """
    for (part_info, part_content) in parts:
        yield (part_info, decodePart(part_info.getencoding(), part_content))

def decodeSignedParts(parts):
    """Decodes bodies of parts to the plain data. Required for processing
       early 8.0.0 backups, which have base64-encoded parts.
    """
    for (part_info, part_content, part_signed) in parts:
        yield (part_info, decodePart(part_info.getencoding(), part_content), part_signed)

def getBundlesList(parts, session_dir):
    """Generates list of bundles. Each bundle must be fixated
       or processed before obtaining subsequent bundle.
    """
    for (partInfo, partContent, signed) in parts:
        contentId = partInfo.getheader("content-id", "")
        b = Bundle(fh = partContent, session_dir = session_dir)
        yield (contentId, b)
        b.invalidate()


schemas = {}

def backupFileFactory(schema, filename, **params):
    for schema_, fileType in schemas.items():
        if schema == schema_:
            try:
                filename = filename.encode("UTF-8")
            except:
                # already encoded
                pass
            return fileType(filename, **params)


class LocalBackupFile(object):
    def __init__(self, filename):
        self.filename = filename

    def exists(self):
        return os.path.exists(self.filename) and os.path.isfile(self.filename)

    def size(self):
        info = os.lstat(self.filename)
        return info[6]

    def time(self):
        info = os.lstat(self.filename)
        return info[8]

    def fh(self, mode='rb'):
        return file(self.filename, mode)

schemas['file'] = LocalBackupFile

class BackupPartsFinder(object):
    """
    Returns *BackupFile-class objects until there are no more backup parts.
    Then returns None
    """
    def __init__(self, schema, filename, **params):
        self.part = filename
        self.schema = schema
        self.params = params
        self.i = 0

    def __call__(self):
        if not self.part:
            return None

        fileName = self.part

        if self.i != 0:
            fileName += ".%03d" % self.i

        try:
            self.i += 1
            return backupFileFactory(self.schema, fileName, **self.params)
        except BackupFailedException, e:
            raise
        except Exception:
            self.part = None
            return None

class Backup(object):
    def __init__(self, schema, filename, session_dir = None, **params):
        self.schema = schema
        self.filename = filename
        self.session_dir = session_dir
        self.params = params

        self.decodedParts = None
        self.metadata = None
        self.dataRead = False
        self.fh = None
        self.signed = False

    def _openFh2(self):
        finder = BackupPartsFinder(self.schema, self.filename, **self.params)
        return FileAdapter(split_files.SplitFile(finder))

    def _openFh(self):
        fh = self._openFh2()
        # remember file object to be able to close it
        self.fh = fh
        if gzip_stream.isGzipped(fh):
            fh = FileAdapter(gzip_stream.GzipReadStream(fh))
        return fh

    def _open(self):
        self.decodedParts = decodeSignedParts(enumerateSignedParts(self._openFh(), True))

    def getMetadata(self):
        if self.metadata:
            raise RuntimeError("Trying to read already read metadata")
        self._open()
        (part_info, part_content, self.signed) = self.decodedParts.next()
        self.metadata = Metadata(part_content)
        return self.metadata

    def isSigned(self):
        return self.signed

    #
    # Returns the list of pairs (content-id, bundle). Metadata is not included
    #
    def getContents(self):
        if self.dataRead:
            raise RuntimeError("Trying to read already read data")
        if not self.metadata:
            self.getMetadata()
        self.metadata.invalidate()
        self.dataRead = True
        return getBundlesList(self.decodedParts, self.session_dir)

    def close(self):
        if self.fh != None:
            self.fh.close()

# Local Variables:
# mode: python
# indent-tabs-mode: nil
# tab-width: 4
# End:
