# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
import commands
import getopt
import codecs
import filters
import os
import osutil
import sys
import backup
import backup_walker
import dump_walker

class ExtractDumpVisitor(object):
    def __init__(self, outDir, verbosity):
        self.outDir = outDir
        self.verbose = verbosity
        self.objects = {}
        self.infoObjects = []
        self.cids = {}

    def log(self, message):
        if self.verbose:
            print message

    def visit(self, info, x, parents):
         baseDir = ""

         if len(parents) > 0:
             parent = parents[-1]
             if self.objects.has_key((parent.type_, parent.name)):
                 self.log("[" + info.__repr__() + "] < [" + parent.__repr__() + "]")
                 baseDir = self.objects[(parent.type_, parent.name)] + "/"

         dir = baseDir + info.objectDir()

         self.log("[" + info.__repr__() + "] -planned-> " + dir)

         self.objects[(info.type_, info.name)] = dir

         for cid, fileName in info.getCids().items():
             self.log(cid + " ~~> " + os.path.join(dir, fileName))
             self.cids[cid] = os.path.join(dir, fileName)

         self.infoObjects.append((dir, x))

    def writeMetadata(self):
        for dir, node in self.infoObjects:
            targetDir = os.path.join(self.outDir, dir)
            self.log(">> %s" % targetDir)
            os.makedirs(targetDir)
            f = file(os.path.join(targetDir, "description.xml"), "wb")
            node.writexml(codecs.lookup("UTF-8")[3](f))
            f.close()

    def unpackCid(self, cid, bundle):
        if cid in self.cids:
            outFile = os.path.join(self.outDir, self.cids[cid])
            del self.cids[cid] # do not use method .pop which is only in python 2.3 and later
            self.log(cid + " -> " + outFile)
            bundle.process(osutil.createToFileSaver(outFile))

            # Stop reading if we read all we need
            if len(self.cids) == 0:
                raise backup_walker.StopProcessing()
        else:
            self.log("IGN " + cid)

class ExtractBackupVisitor(object):
    def __init__(self, filter, outDir, verbosity):
        self.filter = filter
        self.verbose = verbosity
        self.extractor = ExtractDumpVisitor(outDir, self.verbose)

    def visitMetadata(self, metadata):

        def processMetadata(fh):
            dump_walker.walkDump(filters.FilterVisitor(self.extractor, self.filter, self.verbose), fh)
        metadata.process(processMetadata)

        self.extractor.writeMetadata()

    def visitBundle(self, contentId, bundle):
        self.extractor.unpackCid(contentId, bundle)

class ExtractCommand(commands.HelpProvider):
    USAGE_STRING = """extract [-n|--no-recursive]
                     [-d|--directory=<output-directory>] [objects] <backup-file>"""
    SHORT_HELP_STRING = "Extracts the content of backup"
    HELP_STRING = """Extracts the content of backup, or selected parts of backup.

Selecting parts of backup is described in 'backup-unpack help select-objects'.
By default 'extract' extracts objects recursively. This may be disabled by
supplying -n|--no-recursive option.

-d|--directory option explicitly sets the directory for extracting the backup
content. By default directory named after backup file will be created in current
directory.
"""

    def __init__(self, verbosity):
        self.verbosity = verbosity

    def parse(self, args):
        if len(args) == 0:
            raise RuntimeError("No backup file specified")
        self.backupFile = args.pop()

        recursive = True
        self.outDir = os.path.basename(self.backupFile) + ".unpacked/"

        opts, args = getopt.getopt(args, "nd:", ["no-recursive", "directory="])
        for name, value in opts:
            if name in ("-n", "--no-recursive"):
                recursive = False
            if name in ("-d", "--directory"):
                self.outDir = value

        if not self.outDir.endswith("/"):
            self.outDir += "/"

        self.filter = filters.parseStringFilter(args)
        if recursive:
            self.filter.setRecursive()

    def act(self):
        backupFile = backup.backupFileFactory('file', self.backupFile)
        if not backupFile.exists():
            raise RuntimeError("Backup file '" + self.backupFile + "' does not exist")

        if os.path.exists(self.outDir):
            if not os.path.isdir(self.outDir):
                raise RuntimeError(self.outDir + ": Not a directory")
        else:
            os.mkdir(self.outDir)
        
        b = backup.Backup('file', self.backupFile)
        ebv = ExtractBackupVisitor(self.filter, self.outDir, self.verbosity)
        backup_walker.walkBackup(b, ebv)

commands.register("extract", ExtractCommand, 2)

# Local Variables:
# mode: python
# indent-tabs-mode: nil
# tab-width: 4
# End:
