# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
import sys
"""
Filters the supplied objects to select only requested by user
"""

class FilterVisitor(object):
    """Filters the visited objects according to given filter"""
    def __init__(self, visitor, filter, verbose):
        self.visitor = visitor
        self.filter = filter
        self.verbose = verbose
    def visit(self, info, x, parents):
        if self.filter.match(info, parents, self.verbose):
            self.visitor.visit(info, x, parents)


class _ObjectFilter(object):
    def __init__(self):
        self.recursive = False
    def setRecursive(self):
        self.recursive = True
    def match(self, info, parents, verbose):
        if self.recursive and len(parents) > 0 and self.match(parents[0], parents[1:], verbose):
            return True
        return self.test(info, parents, verbose)


class _AllFilter(_ObjectFilter):
    def test(self, info, parents, verbose):
        return True
    def __repr__(self):
        return "<#All>"

class _SingleObjectFilter(_ObjectFilter):
    def __init__(self, type_, name):
        _ObjectFilter.__init__(self)
        self.type_ = type_
        self.name = name
    def test(self, info, parents, verbose):
        if verbose:
            print "Matching %s %s with %s %s" % (self.type_, self.name, info.type_, info.name)
        return info.type_ == self.type_ and info.name == self.name
    def __repr__(self):
        return "<#SingleObject: %s: %s>" % (self.name, self.type_)


class _ObjectTypeFilter(_ObjectFilter):
    def __init__(self, type_):
        _ObjectFilter.__init__(self)
        self.type_ = type_
    def test(self, info, parents, verbose):
        return info.type_ == self.type_
    def __repr__(self):
        return "<#ObjectType: %s>" % self.type_


class _SingleObjectWithParentFilter(_ObjectFilter):
    def __init__(self, type_, name, parentName):
        _ObjectFilter.__init__(self)
        self.type_ = type_
        self.name = name
        self.parentName = parentName
    def test(self, info, parents, verbose):
        return info.type_ == self.type_ and info.name == self.name and \
            len(parents) > 0 and parents[-1].name == self.parentName
    def __repr__(self):
        return "<#SingleObjectWithParent: %s: %s under %s>" % (self.name, self.type_, self.parentName)

class _DatabaseFilter(_ObjectFilter):
    def __init__(self, name, host, port, type_):
        self.name = name
        self.host = host
        self.port = port
        self.type_ = type_

    def test(self, info, parents, verbose):
        if info.type_ != "database": return False
        if self.type_ and info.dbType != self.type_: return False
        if self.host and info.host != self.host: return False
        if self.port and info.port != self.port: return False
        return info.name != self.name

    def __repr__(self):
        return "<#Database: %s@%s:%s/%s>" % (self.name, self.host or "",
                                             self.port or "", self.type_ or "")


def parseStringFilter(args):
    if len(args) == 0:
        return _AllFilter()
    if len(args) == 1:
        # multiple entities or short selection
        arg = args[0]
        if arg in ["clients", "domains", "mailnames", "maillists", "databases", "webusers"]:
            # plural -> singular
            return _ObjectTypeFilter(_toSingular(arg))
        if _isLikeEmail(arg):
            (mailname, domain) = arg.split("@", 1)
            return _SingleObjectWithParentFilter("mailname", mailname, domain)
        if _isLikeDomain(arg):
            return _SingleObjectFilter("domain", arg)
        if _isLikeClient(arg):
            return _SingleObjectFilter("client", arg)
        raise RuntimeError("Unknown object: " + arg)
    if len(args) == 2:
        #full selection
        (type_, name) = args
        if type_ in ["client", "domain"]:
            return _SingleObjectFilter(type_, name)
        if type_ in ["mailname", "maillist", "webuser"]:
            (object, parent) = _parseObjectWithParent(name)
            return _SingleObjectWithParentFilter(type_, object, parent)
        if type_ == "database":
            return _parseDatabase(name)
        raise RuntimeError("Unknown object type: " + type_)
    raise RuntimeError("Too many arguments")

def _isLikeEmail(s):
    return "@" in s

def _isLikeDomain(s):
    return "." in s

def _isLikeClient(s):
    return True

def _parseObjectWithParent(s):
    if not "@" in s:
        raise RuntimeError("Unable to parse object name: " + s)
    return s.split("@", 2)

def _parseDatabase(s):
    # Type
    if s.endswith("/mysql"):
        type_ = "mysql"
        s = s[:-6]
    elif s.endswith("/postgresql"):
        type_ = "postgresql"
        s = s[:-11]
    else:
        type_ = None

    # DB server
    if "@" in s:
        host = s[s.rfind("@") + 1:]
        s = s[:s.rfind("@")]

        if ":" in host:
            try:
                port = host[host.rfind(":") + 1:]
                port = int(port)
            except ValueError:
                raise RuntimeError("Unable to parse database server port: " + port)
            host = host[:host.rfind(":")]
        else:
            port = None
    else:
        host = None
        port = None

    return _DatabaseFilter(s, host, port, type_)

# Works only for plural form of entity types
def _toSingular(s):
    return s[:-1]

# Local Variables:
# mode: python
# indent-tabs-mode: nil
# tab-width: 4
# End:
