#!/usr/bin/env python
# Copyright 1999-2017. Plesk International GmbH. All rights reserved.

import os
import sys
from optparse import OptionParser
import logging as log
import fileinput
import plesk_subprocess as subprocess
import re

PLESK_LIBEXEC_DIR = '/usr/lib/plesk-9.0'

# http://en.wikipedia.org/wiki/Hostname#Restrictions_on_valid_host_names
def valid_hostname(hostname):
    """ Check that provided hostname is valid. """
    if len(hostname) > 255:
        return False
    if hostname.endswith("."): # A single trailing dot is legal
        hostname = hostname[:-1] # strip exactly one dot from the right, if present
    disallowed = re.compile("[^A-Z0-9-]", re.IGNORECASE) # The RFC mandate that component hostname labels may contain only the ASCII letters 'a' through 'z' (in a case-insensitive manner), the digits '0' through '9', and the hyphen ('-').
    return all(  # Split by labels and verify individually
        (label and len(label) <= 63 and  # length is within proper range
         not label.startswith("-") and not label.endswith("-") and  # no bordering hyphens
         not disallowed.search(label))  # contains only legal characters
        for label in hostname.split("."))

def update_hostname_file(filename, fqdn):
    """ Function to write new hostname into files like /etc/hostname, /etc/mailname and so on """
    if not os.path.isfile(filename):
        log.debug("%s not updated, as it's not regular file", filename)
        return False

    with open(filename, 'w') as f:
        f.write("%s\n" % fqdn)

def _extract_short_hostname(canonical_hostname):
    if isinstance(canonical_hostname, basestring):
        return canonical_hostname.split('.')[0]

    return None

def update_hosts_file(filename, old_fqdn, new_fqdn, main_ip):
    """ Update hosts file by replacing old hostnames with new ones and fixing important entries """

    CANONICAL_LOCALHOST = "localhost.localdomain"
    SHORT_LOCALHOST = _extract_short_hostname(CANONICAL_LOCALHOST)
    LOOPBACK_ADDRESSES = ["127.0.0.1", "::1"]

    CANONICAL_LOCALHOST_NAMES = [CANONICAL_LOCALHOST, "localhost4.localdomain4", "localhost6.localdomain6"]
    SHORT_LOCALHOST_NAMES = [_extract_short_hostname(x) for x in CANONICAL_LOCALHOST_NAMES]

    old_short_name = _extract_short_hostname(old_fqdn)
    new_short_name = _extract_short_hostname(new_fqdn)

    if new_fqdn == CANONICAL_LOCALHOST and main_ip not in LOOPBACK_ADDRESSES:
        main_ip = LOOPBACK_ADDRESSES[0]
    elif new_fqdn == "localhost4.localdomain4":
        main_ip = "127.0.0.1"
    elif new_fqdn == "localhost6.localdomain6":
        main_ip = "::1"

    # The hosts file is a simple text file that associates IP addresses with hostnames
    with open(filename, 'r+') as f:
        lines = f.readlines()

        # The file is rewritten on every call
        f.seek(0)
        f.truncate()

        invalid_entries_fixed = False
        hostname_updated = False

        for line in lines:
            entry = line.strip()
            comment = ""

            # Separate a comment
            comment_pos = entry.find('#')
            if comment_pos != -1:
                comment = entry[comment_pos:]
                entry = entry[:comment_pos].rstrip()

            # Skip an empty entry
            if not entry:
                f.write(line)
                continue

            data = entry.split()

            # Comment out an entry if it does not conform to the format of the hosts file
            if len(data) < 2:
                f.write('#' + line)
                invalid_entries_fixed = True
                continue

            ip_address = data[0]
            hostnames = data[1:]

            if new_fqdn in CANONICAL_LOCALHOST_NAMES:
                hostnames = filter(lambda x: x not in [old_fqdn, old_short_name], hostnames)
            elif ip_address == main_ip:
                [hostnames.insert(0, x) for x in [new_short_name, new_fqdn]]
            else:
                hostnames = filter(lambda x: x not in [new_fqdn, new_short_name], hostnames)

            if old_fqdn in CANONICAL_LOCALHOST_NAMES and ip_address in LOOPBACK_ADDRESSES:
                # Keep those localhosts that are associated with a loopback address unchanged
                pass
            elif old_fqdn in hostnames:
                hostnames = filter(lambda x: x not in [old_fqdn, old_short_name], hostnames)
                [hostnames.insert(0, x) for x in [new_short_name, new_fqdn]]

            # Make localhost.localdomain resolve into a loopback address
            if ip_address in LOOPBACK_ADDRESSES:
                if ip_address == main_ip:
                    [hostnames.insert(0, x) for x in [new_short_name, new_fqdn]]
                [hostnames.insert(0, x) for x in [SHORT_LOCALHOST, CANONICAL_LOCALHOST]]
            else:
                hostnames = filter(lambda x: x not in CANONICAL_LOCALHOST_NAMES and x not in SHORT_LOCALHOST_NAMES, hostnames)

            # Take off duplicate hostnames
            unique_hostnames = []
            [unique_hostnames.append(x) for x in hostnames if x not in unique_hostnames]

            # Write an updated entry back to the file
            if len(unique_hostnames) > 0:
                f.write("%s\t%s\t%s\n" % (ip_address, " ".join(unique_hostnames), comment))

            if ip_address == main_ip and new_fqdn in unique_hostnames:
                hostname_updated = True

        if not hostname_updated:
            f.write("\n%s\t%s %s" % (main_ip, new_fqdn, new_short_name))
            log.info("%s has been associated with %s" % (new_fqdn, main_ip))

        if invalid_entries_fixed:
            log.warn("The entries that do not conform to the format of the hosts file have been commented out")

def update_network_file(filename, new_fqdn):
    """ Replace hostname in /etc/sysconfig/network file """
    if not os.path.isfile(filename):
        log.debug("Do not update '%s', as it's not regular file", filename)
        return

    for line in fileinput.input(filename, inplace=True):
        if line.strip().startswith("HOSTNAME="):
            sys.stdout.write("HOSTNAME=%s\n" % new_fqdn)
        else:
            sys.stdout.write(line)

def get_hostname():
    """ Obtain current hostname from system """
    try:
        return subprocess.check_output(["/bin/hostname", "--fqdn"])
    except subprocess.CalledProcessError:
        log.debug("Cannot get FQDN hostname")

    return None

def set_hostname(hostname, domainname, main_ip):
    """ Set hostname on Linux host """
    new_fqdn = "%s.%s" % (hostname, domainname)

    old_fqdn = get_hostname()
    if type(old_fqdn) is str:
        old_fqdn = old_fqdn.strip()
        log.debug("Current hostname is %s", old_fqdn)

    if new_fqdn == old_fqdn:
        return

    log.debug("Update hostname in system")
    subprocess.check_call(["/bin/hostname", new_fqdn])

    for hostname_file in ["/etc/hostname", "/etc/HOSTNAME"]:
        log.debug("Update hostname in %s, if exists", hostname_file)
        update_hostname_file(hostname_file, new_fqdn)

    log.debug("Update hostname in /etc/hosts")
    update_hosts_file("/etc/hosts", old_fqdn, new_fqdn, main_ip)

    log.debug("Update hostname in /etc/sysconfig/network file")
    update_network_file("/etc/sysconfig/network", new_fqdn)

    if os.path.isfile(os.path.join(PLESK_LIBEXEC_DIR, "mailsrv_set_hostname")):
        log.debug("Update hostname via mailsrv_set_hostname")
        subprocess.check_call([os.path.join(PLESK_LIBEXEC_DIR, "mailsrv_set_hostname"), new_fqdn])

    log.debug("Update hostname in /etc/mailname")
    update_hostname_file("/etc/mailname", new_fqdn)

def reconfig():
    """ Parse command line arguments and do the work """
    usage = "usage: %prog [-v] <hostname> <domain-name> <main-ip>"
    parser = OptionParser(usage=usage)
    parser.description = "Utility set hostname"
    parser.add_option('-v', '--verbose', action='count', dest='verbose',
                      help='Increase verbosity (can be specified multiple times)')

    (options, args) = parser.parse_args()
    log_level = log.WARNING

    if options.verbose == 1:
        log_level = log.INFO
    elif options.verbose >= 2:
        log_level = log.DEBUG

    log.basicConfig(format='%(levelname)s:%(message)s', level=log_level)

    log.debug("Options: %s", options)
    log.debug("Args: %s", args)

    if len(args) < 2:
        parser.error("Not enough arguments")

    hostname = args[0]
    domainname = args[1]
    main_ip = args[2] if len(args) > 2 else "127.0.0.1" # main_ip is an optional argument (temporarily)

    if not hostname or not domainname or not main_ip:
        parser.error("Not enough arguments")

    if not valid_hostname("%s.%s" % (hostname, domainname)):
        parser.error("Specified hostname is invalid")

    if os.geteuid() != 0:
        raise RuntimeError("You need to have root privileges to run this utility")

    return set_hostname(hostname, domainname, main_ip)

def main():
    """ reconfig main entry point """
    try:
        return reconfig()
    except SystemExit:
        raise
    except Exception, ex:
        sys.stderr.write('%s\n' % ex)
        log.debug('This exception happened at:', exc_info=sys.exc_info())
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :

